#include <opencv2/opencv.hpp>
#include <fstream>
#include <string>

// ւ̕ϊ֐
template<typename T>
std::string to_string(const T &value)
{
  std::ostringstream stringStream;
  stringStream << value;
  return stringStream.str();
}

// AvP[VNX
class Application {
public:

  // RXgN^
  Application();

  // fXgN^
  ~Application();

  // C[v
  void run();

private:

  // [h
  enum ProcessingMode {
    ProcessingMode_Run,
    ProcessingMode_OriginalView,
    ProcessingMode_MedianSetting,
    ProcessingMode_GaussianSetting,
    ProcessingMode_BlackhatSetting,
    ProcessingMode_BlackhatGainSetting,
    ProcessingMode_NoiseSuppressionSetting,
    ProcessingMode_CalibrationStandby,
    ProcessingMode_CalibrationStep1,
    ProcessingMode_CalibrationStep2,
    ProcessingMode_CalibrationStep3,
    ProcessingMode_CalibrationStep4
  };

  // rfILv`[foCXԍ
  const int _videoCaptureDeviceId;

  // CEBhE
  const char *_mainWindowName;

  // fBXvC̃TCY
  const cv::Size _displayImageSize;

  // p[^t@C
  const char *_parameterFileName;

  // rfILv`[IuWFNg
  cv::VideoCapture _videoCapture;

  // J摜TCY
  cv::Size _cameraImageSize;
  
  // ROI
  cv::Rect _roi;

  // 摜
  cv::Mat _inputImage;
  cv::Mat _grayImage;
  cv::Mat _medianImage;
  cv::Mat _gaussianImage;
  cv::Mat _blackhatImage;
  cv::Mat _noiseSuppressionImage;
  cv::Mat _resultImage;
  cv::Mat _displayImage;

  // fBAtB^[TCY
  int _medianFilterSize;

  // KEVAtB^[TCY
  int _gaussianFilterSize;

  // ubNnbg\vfTCY
  int _blackhatStructuringElementSize;

  // ubNnbgQC
  int _blackhatGain;

  // mCY}臒l
  int _noiseSuppressionThreshold;

  // Lu[V|CgԊu
  int _calibrationPointDistance;

  // JWñLu[V|Cg
  cv::Point2f _cameraCalibrationPoints[4];

  // fBXvCW̃Lu[V|Cg
  cv::Point2f _displayCalibrationPoints[4];

  // JWnfBXvCWnւ̕ϊs
  cv::Mat _transformMatrix;

  // [h
  ProcessingMode _processingMode;
  
  // mCY}
  static void suppressNoise(const cv::Mat &inputImage, cv::Mat &outputImage, int threshold);

  // CEBhẼ}EXR[obN֐
  static void mainWindowMouseCallback(int event, int x, int y, int flags, void *userdata);

  // CEBhẼ}EXCxgnh
  void mainWindowOnMouseEvent(int event, int x, int y, int flags);

  // L[Cxg
  bool processKey(int key);

  // ROI̍XV
  void updateRoi();

  // fBXvCW̃Lu[V|Cg̍XV
  void updateDisplayCalibrationPoints();

};

// RXgN^
Application::Application()
  : _videoCaptureDeviceId(0),
    _mainWindowName("Blood Vessel Detection"),
    _displayImageSize(1182, 556),
    _parameterFileName("BloodVesselDetection.dat"),
    _videoCapture(),
    _cameraImageSize(),
    _roi(),
    _inputImage(),
    _grayImage(),
    _medianImage(),
    _gaussianImage(),
    _blackhatImage(),
    _noiseSuppressionImage(),
    _resultImage(),
    _displayImage(_displayImageSize, CV_8UC3),
    _medianFilterSize(5),
    _gaussianFilterSize(11),
    _blackhatStructuringElementSize(1),
    _blackhatGain(10),
    _noiseSuppressionThreshold(30),
    _calibrationPointDistance(7),
    _cameraCalibrationPoints(),
    _displayCalibrationPoints(),
    _transformMatrix(),
    _processingMode(ProcessingMode_Run)
{
  // rfILv`foCXI[v
  _videoCapture.open(_videoCaptureDeviceId);

  // 摜t[TCY̎擾
  _cameraImageSize = cv::Size(static_cast<int>(_videoCapture.get(cv::CAP_PROP_FRAME_WIDTH)),
                              static_cast<int>(_videoCapture.get(cv::CAP_PROP_FRAME_HEIGHT)));

  // 摜
  _resultImage.create(_cameraImageSize, CV_8UC3);
  _resultImage = 0;

  // JWñLu[V|Cg̏l
  _cameraCalibrationPoints[0] = cv::Point2f(0.1f * _cameraImageSize.width, 0.1f * _cameraImageSize.height);
  _cameraCalibrationPoints[1] = cv::Point2f(0.9f * _cameraImageSize.width, 0.1f * _cameraImageSize.height);
  _cameraCalibrationPoints[2] = cv::Point2f(0.9f * _cameraImageSize.width, 0.9f * _cameraImageSize.height);
  _cameraCalibrationPoints[3] = cv::Point2f(0.1f * _cameraImageSize.width, 0.9f * _cameraImageSize.height);

  // ROI̍XV
  updateRoi();

  // fBXvCWñLu[V|CgXV
  updateDisplayCalibrationPoints();

  // ftHg̍Wϊs̎Zo
  _transformMatrix = cv::getPerspectiveTransform(_cameraCalibrationPoints, _displayCalibrationPoints);

  // p[^̃[h
  std::ifstream file(_parameterFileName);
  if (file) {
    file >> _roi.x
         >> _roi.y
         >> _roi.width
         >> _roi.height;
    file >> _medianFilterSize;
    file >> _gaussianFilterSize;
    file >> _blackhatStructuringElementSize;
    file >> _blackhatGain;
    file >> _noiseSuppressionThreshold;
    file >> _calibrationPointDistance;
    file >> _transformMatrix.at<double>(0, 0)
         >> _transformMatrix.at<double>(0, 1)
         >> _transformMatrix.at<double>(0, 2)
         >> _transformMatrix.at<double>(1, 0)
         >> _transformMatrix.at<double>(1, 1)
         >> _transformMatrix.at<double>(1, 2)
         >> _transformMatrix.at<double>(2, 0)
         >> _transformMatrix.at<double>(2, 1)
         >> _transformMatrix.at<double>(2, 2);
    updateDisplayCalibrationPoints();
  }
}

// fXgN^
Application::~Application()
{
  // p[^̃Z[u
  std::ofstream file(_parameterFileName);
  file << _roi.x << "\t"
       << _roi.y << "\t"
       << _roi.width << "\t"
       << _roi.height << std::endl;
  file << _medianFilterSize << std::endl;
  file << _gaussianFilterSize << std::endl;
  file << _blackhatStructuringElementSize << std::endl;
  file << _blackhatGain << std::endl;
  file << _noiseSuppressionThreshold << std::endl;
  file << _calibrationPointDistance << std::endl;
  file << _transformMatrix.at<double>(0, 0) << "\t"
       << _transformMatrix.at<double>(0, 1) << "\t"
       << _transformMatrix.at<double>(0, 2) << "\t"
       << _transformMatrix.at<double>(1, 0) << "\t"
       << _transformMatrix.at<double>(1, 1) << "\t"
       << _transformMatrix.at<double>(1, 2) << "\t"
       << _transformMatrix.at<double>(2, 0) << "\t"
       << _transformMatrix.at<double>(2, 1) << "\t"
       << _transformMatrix.at<double>(2, 2) << std::endl;
}

// C[v
void Application::run()
{
  // CEBhE̐
  cv::namedWindow(_mainWindowName);
  cv::moveWindow(_mainWindowName, 0, 0);
  cv::setMouseCallback(_mainWindowName, mainWindowMouseCallback, this);

  cv::Mat calibrationImage;

  while (true) {
    // 1t[荞
    _videoCapture >> _inputImage;

    // N[摜ɕϊqXgOϓ
    cv::cvtColor(_inputImage(_roi), _grayImage, CV_BGR2GRAY);
    cv::equalizeHist(_grayImage, _grayImage);

    // fBAtB^[
    if (1 < _medianFilterSize) {
      cv::medianBlur(_grayImage, _medianImage, _medianFilterSize);
    }
    else {
      _grayImage.copyTo(_medianImage);
    }

    // KEVAtB^[
    if (1 < _gaussianFilterSize) {
      cv::GaussianBlur(_medianImage,
                       _gaussianImage,
                       cv::Size(_gaussianFilterSize, _gaussianFilterSize),
                       0.0);
    }
    else {
      _medianImage.copyTo(_gaussianImage);
    }

    // ubNnbgϊ
    if (1 < _blackhatStructuringElementSize) {
      cv::morphologyEx(_gaussianImage,
                       _blackhatImage,
                       cv::MORPH_BLACKHAT,
                       cv::getStructuringElement(cv::MORPH_ELLIPSE,
                                                 cv::Size(_blackhatStructuringElementSize,
                                                          _blackhatStructuringElementSize)));
      _blackhatImage.convertTo(_blackhatImage, -1, double(_blackhatGain) / 10.0);
    }
    else {
      _blackhatImage = ~_gaussianImage;
    }

    // mCY}
    suppressNoise(_blackhatImage, _noiseSuppressionImage, _noiseSuppressionThreshold);

    // [hɂĈقȂ鏈
    switch (_processingMode) {
    case ProcessingMode_Run: {
      // Fʉ摜쐬
      int channelMap[2] = {
        0,
        1
      };
      cv::mixChannels(_noiseSuppressionImage, _resultImage(_roi), channelMap, 1);

      // ʉ摜JWfBXvCW֕ϊĕ\
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      cv::imshow(_mainWindowName, _displayImage);
      break; }
    case ProcessingMode_OriginalView:
      // ͉摜JWfBXvCW֕ϊĕ\
      cv::warpPerspective(_inputImage, _displayImage, _transformMatrix, _displayImageSize);
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_MedianSetting:
      // fBA摜JWfBXvCW֕ϊĕ\
      cv::cvtColor(_medianImage, _resultImage(_roi), CV_GRAY2BGR);
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      if (1 < _medianFilterSize) {
        cv::putText(_displayImage, "Median Filter: " + to_string(_medianFilterSize), cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      else {
        cv::putText(_displayImage, "Median Filter: off", cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_GaussianSetting:
      // KEVA摜JWfBXvCW֕ϊĕ\
      cv::cvtColor(_gaussianImage, _resultImage(_roi), CV_GRAY2BGR);
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      if (1 < _gaussianFilterSize) {
        cv::putText(_displayImage, "Gaussian Filter: " + to_string(_gaussianFilterSize), cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      else {
        cv::putText(_displayImage, "Gaussian Filter: off", cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_BlackhatSetting:
      // ubNnbg摜JWfBXvCW֕ϊĕ\
      cv::cvtColor(_blackhatImage, _resultImage(_roi), CV_GRAY2BGR);
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      if (1 < _blackhatStructuringElementSize) {
        cv::putText(_displayImage, "Blackhat Transform: " + to_string(_blackhatStructuringElementSize), cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      else {
        cv::putText(_displayImage, "Blackhat Transform: off", cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_BlackhatGainSetting:
      // ubNnbg摜JWfBXvCW֕ϊĕ\
      cv::cvtColor(_blackhatImage, _resultImage(_roi), CV_GRAY2BGR);
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      cv::putText(_displayImage, "Blackhat Gain: " + to_string(double(_blackhatGain) / 10.0).substr(0, 3), cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_NoiseSuppressionSetting:
      // ubNnbg摜JWfBXvCW֕ϊĕ\
      cv::cvtColor(_noiseSuppressionImage, _resultImage(_roi), CV_GRAY2BGR);
      cv::warpPerspective(_resultImage, _displayImage, _transformMatrix, _displayImageSize);
      if (_noiseSuppressionThreshold) {
        cv::putText(_displayImage, "Noise Suppression: " + to_string(_noiseSuppressionThreshold), cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      else {
        cv::putText(_displayImage, "Noise Suppression: off", cv::Point(10, 40), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0, 0, 255), 2);
      }
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_CalibrationStandby:
      _inputImage.copyTo(calibrationImage);
      _displayImage = 0;
      for (int i = 0; i < 4; ++i) {
        cv::line(_displayImage,
                 cv::Point(_displayCalibrationPoints[i]) - cv::Point(30, 0),
                 cv::Point(_displayCalibrationPoints[i]) + cv::Point(30, 0),
                 cv::Scalar(0, 0, 255),
                 3);
        cv::line(_displayImage,
                 cv::Point(_displayCalibrationPoints[i]) - cv::Point(0, 30),
                 cv::Point(_displayCalibrationPoints[i]) + cv::Point(0, 30),
                 cv::Scalar(0, 0, 255),
                 3);
      }
      cv::imshow(_mainWindowName, _displayImage);
      break;
    case ProcessingMode_CalibrationStep1:
    case ProcessingMode_CalibrationStep2:
    case ProcessingMode_CalibrationStep3:
    case ProcessingMode_CalibrationStep4:
      cv::imshow(_mainWindowName, calibrationImage);
      break;
    }

    // EFCgƃL[Cxg̏
    if (!processKey(cv::waitKey(33))) {
      break;
    }
  }

  // CEBhE̔j
  cv::destroyWindow(_mainWindowName);
}

// mCY}
void Application::suppressNoise(const cv::Mat &inputImage, cv::Mat &outputImage, int threshold)
{
  if (threshold) {
    cv::Size imageSize = inputImage.size();
    outputImage.create(imageSize, CV_8UC1);
    cv::Mat_<unsigned char> input(inputImage);
    cv::Mat_<unsigned char> output(outputImage);
    for (int y = 0; y < imageSize.height; ++y) {
      const unsigned char *inputLine = input[y];
      unsigned char *outputLine = output[y];
      for (int x = 0; x < imageSize.width; ++x) {
        outputLine[x] = inputLine[x] < threshold ? 0 : inputLine[x];
      }
    }
  }
  else {
    inputImage.copyTo(outputImage);
  }
}

// CEBhẼ}EXR[obN
void Application::mainWindowMouseCallback(int event, int x, int y, int flags, void *userdata)
{
  // ̓CX^X\bh̃CxgnhɈϏ
  Application *application = static_cast<Application *>(userdata);
  if (application) {
    application->mainWindowOnMouseEvent(event, x, y, flags);
  }
}

// CEBhẼ}EXCxgnh
void Application::mainWindowOnMouseEvent(int event, int x, int y, int flags)
{
  switch (event) {
  case cv::EVENT_LBUTTONDOWN:
    switch (_processingMode) {
    case ProcessingMode_CalibrationStep1:
      // 1Ԗڂ̃Lu[V|CgZbg
      _cameraCalibrationPoints[0] = cv::Point2f(static_cast<float>(x), static_cast<float>(y));
      _processingMode = ProcessingMode_CalibrationStep2;
      break;
    case ProcessingMode_CalibrationStep2:
      // 2Ԗڂ̃Lu[V|CgZbg
      _cameraCalibrationPoints[1] = cv::Point2f(static_cast<float>(x), static_cast<float>(y));
      _processingMode = ProcessingMode_CalibrationStep3;
      break;
    case ProcessingMode_CalibrationStep3:
      // 3Ԗڂ̃Lu[V|CgZbg
      _cameraCalibrationPoints[2] = cv::Point2f(static_cast<float>(x), static_cast<float>(y));
      _processingMode = ProcessingMode_CalibrationStep4;
      break;
    case ProcessingMode_CalibrationStep4:
      // 4Ԗڂ̃Lu[V|CgZbgĕϊXV
      _cameraCalibrationPoints[3] = cv::Point2f(static_cast<float>(x), static_cast<float>(y));
      _transformMatrix = cv::getPerspectiveTransform(_cameraCalibrationPoints, _displayCalibrationPoints);

      // ROI̍XV
      updateRoi();
      _processingMode = ProcessingMode_Run;
      _resultImage = 0;
      break;
    }
  }
}

// L[Cxg
bool Application::processKey(int key)
{
  // AvP[VpꍇtrueԂ
  bool result = true;

  switch (key) {
  case 's': // 摜ۑ
    cv::imwrite("InputImage.png", _inputImage);
    cv::imwrite("GrayImage.png", _grayImage);
    cv::imwrite("MedianImage.png", _medianImage);
    cv::imwrite("GaussianImage.png", _gaussianImage);
    cv::imwrite("BlackhatImage.png", _blackhatImage);
    cv::imwrite("NoiseSuppressionImage.png", _noiseSuppressionImage);
    break;
  case 0x0000000a: // Enter
    if (_processingMode == ProcessingMode_CalibrationStandby) {
      _processingMode = ProcessingMode_CalibrationStep1;
    }
    break;
  case 0x0000001b: // ESC
    result = false;
    break;
  case 0x0000ff51: // 
    // ݒ胂[h̏ꍇAl
    switch (_processingMode) {
    case ProcessingMode_MedianSetting:
      _medianFilterSize = std::max(_medianFilterSize - 2, 1);
      break;
    case ProcessingMode_GaussianSetting:
      _gaussianFilterSize = std::max(_gaussianFilterSize - 2, 1);
      break;
    case ProcessingMode_BlackhatSetting:
      _blackhatStructuringElementSize = std::max(_blackhatStructuringElementSize - 2, 1);
      break;
    case ProcessingMode_BlackhatGainSetting:
      _blackhatGain = std::max(_blackhatGain - 1, 1);
      break;
    case ProcessingMode_NoiseSuppressionSetting:
      _noiseSuppressionThreshold = std::max(_noiseSuppressionThreshold - 1, 0);
      break;
    case ProcessingMode_CalibrationStandby:
      _calibrationPointDistance = std::max(_calibrationPointDistance - 1, 1);
      updateDisplayCalibrationPoints();
      break;
    }
    break;
  case 0x0000ff52: // 
    // t̃[hؑ
    switch (_processingMode) {
    case ProcessingMode_Run:
      _processingMode = ProcessingMode_CalibrationStandby;
      break;
    case ProcessingMode_OriginalView:
      _processingMode = ProcessingMode_Run;
      _resultImage = 0;
      break;
    case ProcessingMode_MedianSetting:
      _processingMode = ProcessingMode_OriginalView;
      break;
    case ProcessingMode_GaussianSetting:
      _processingMode = ProcessingMode_MedianSetting;
      break;
    case ProcessingMode_BlackhatSetting:
      _processingMode = ProcessingMode_GaussianSetting;
      break;
    case ProcessingMode_BlackhatGainSetting:
      _processingMode = ProcessingMode_BlackhatSetting;
      break;
    case ProcessingMode_NoiseSuppressionSetting:
      if (1 < _blackhatStructuringElementSize) {
        _processingMode = ProcessingMode_BlackhatGainSetting;
      }
      else {
        _processingMode = ProcessingMode_BlackhatSetting;
      }
      break;
    default:
      _processingMode = ProcessingMode_NoiseSuppressionSetting;
      break;
    }
    break;
  case 0x0000ff53: // 
    // ݒ胂[h̏ꍇAl𑝉
    switch (_processingMode) {
    case ProcessingMode_MedianSetting:
      _medianFilterSize = std::min(_medianFilterSize + 2, 51);
      break;
    case ProcessingMode_GaussianSetting:
      _gaussianFilterSize = std::min(_gaussianFilterSize + 2, 51);
      break;
    case ProcessingMode_BlackhatSetting:
      _blackhatStructuringElementSize = std::min(_blackhatStructuringElementSize + 2, 51);
      break;
    case ProcessingMode_BlackhatGainSetting:
      _blackhatGain = std::min(_blackhatGain + 1, 50);
      break;
    case ProcessingMode_NoiseSuppressionSetting:
      _noiseSuppressionThreshold = std::min(_noiseSuppressionThreshold + 1, 255);
      break;
    case ProcessingMode_CalibrationStandby:
      _calibrationPointDistance = std::min(_calibrationPointDistance + 1, 9);
      updateDisplayCalibrationPoints();
      break;
    }
    break;
  case 0x0000ff54: // 
    // ̃[hؑ
    switch (_processingMode) {
    case ProcessingMode_Run:
      _processingMode = ProcessingMode_OriginalView;
      break;
    case ProcessingMode_OriginalView:
      _processingMode = ProcessingMode_MedianSetting;
      break;
    case ProcessingMode_MedianSetting:
      _processingMode = ProcessingMode_GaussianSetting;
      break;
    case ProcessingMode_GaussianSetting:
      _processingMode = ProcessingMode_BlackhatSetting;
      break;
    case ProcessingMode_BlackhatSetting:
      if (1 < _blackhatStructuringElementSize) {
        _processingMode = ProcessingMode_BlackhatGainSetting;
      }
      else {
        _processingMode = ProcessingMode_NoiseSuppressionSetting;
      }
      break;
    case ProcessingMode_BlackhatGainSetting:
      _processingMode = ProcessingMode_NoiseSuppressionSetting;
      break;
    case ProcessingMode_NoiseSuppressionSetting:
      _processingMode = ProcessingMode_CalibrationStandby;
      break;
    default:
      _processingMode = ProcessingMode_Run;
      _resultImage = 0;
      break;
    }
    break;
  }

  return result;
}

// ROI̍XV
void Application::updateRoi()
{
  _roi = cv::Rect(cv::Point2f(std::min(std::min(std::min(_cameraCalibrationPoints[0].x,
                                                         _cameraCalibrationPoints[1].x),
                                                _cameraCalibrationPoints[2].x),
                                       _cameraCalibrationPoints[3].x),
                              std::min(std::min(std::min(_cameraCalibrationPoints[0].y,
                                                         _cameraCalibrationPoints[1].y),
                                                _cameraCalibrationPoints[2].y),
                                       _cameraCalibrationPoints[3].y)),
                  cv::Point2f(std::max(std::max(std::max(_cameraCalibrationPoints[0].x,
                                                         _cameraCalibrationPoints[1].x),
                                                _cameraCalibrationPoints[2].x),
                                       _cameraCalibrationPoints[3].x),
                              std::max(std::max(std::max(_cameraCalibrationPoints[0].y,
                                                         _cameraCalibrationPoints[1].y),
                                                _cameraCalibrationPoints[2].y),
                                       _cameraCalibrationPoints[3].y)));
}

// fBXvCW̃Lu[V|Cg̍XV
void Application::updateDisplayCalibrationPoints()
{
  float ratio = static_cast<float>(_calibrationPointDistance) / 20.0f;
  _displayCalibrationPoints[0] = cv::Point2f((0.5f - ratio) * _displayImageSize.width,
                                             (0.5f - ratio) * _displayImageSize.height);
  _displayCalibrationPoints[1] = cv::Point2f((0.5f + ratio) * _displayImageSize.width,
                                             (0.5f - ratio) * _displayImageSize.height);
  _displayCalibrationPoints[2] = cv::Point2f((0.5f + ratio) * _displayImageSize.width,
                                             (0.5f + ratio) * _displayImageSize.height);
  _displayCalibrationPoints[3] = cv::Point2f((0.5f - ratio) * _displayImageSize.width,
                                             (0.5f + ratio) * _displayImageSize.height);
}

// C֐
int main()
{
  // AvP[VIuWFNg̐
  Application application;

  // C[v̎s
  application.run();

  return 0;
}
